//Echoes' fragility system and Abyss object stuff, coded by Emmette Clyde

local SkinCheck = Echoes.SkinCheck
local ObjectSpawner = Echoes.ObjectSpawner
local SpeedCalc = Echoes.SpeedCalc
local StateSetter = Echoes.StateSetter
local IsNiGHTS = Echoes.IsNiGHTS
local JumpManager = Echoes.JumpManager
local ShieldColorFiller = Echoes.ShieldColorFiller
local ParticleSpawner = Echoes.ParticleSpawner
local ObjectZCoord = Echoes.ObjectZCoord
local SearchRadius = Echoes.SearchRadius
local isRing = Echoes.isRing
local isSuper = Echoes.isSuper
local PainSound = Echoes.PainSound

//if echoes takes damage while abyss is active, he'll take damage for her
//unless damage source is elemental and abyss has shield perks
local Immunities = {
	[DMG_FIRE] = SH_PROTECTFIRE,
	[DMG_ELECTRIC] = SH_PROTECTELECTRIC,
	[DMG_WATER] = SH_PROTECTWATER
}

local function FragilityDamage(target, inflictor, source, damage, damagetype)
	if target != nil and target.valid and target.player != nil and 
			target.player.valid and target.skin == "echoes" and not
			IsNiGHTS(target.player)
		//failsafe
		if (damagetype != nil and damagetype & DMG_DEATHMASK) or not 
				G_CoopGametype() or G_IsSpecialStage(gamemap)
			return
		end
		local echoes = target.player.echoes
		if echoes != nil
			if Immunities[damagetype] != nil and echoes.abyss_shield != 
					SH_NONE and echoes.abyss_shield & Immunities[damagetype]
				return true
			end
			if echoes.voidtumble
				echoes.voidtumble = false
				target.spriteroll = 0
				target.player.powers[pw_strong] = STR_NONE
			end
			//if echoes is void gliding or phantom dashing, knockback instead
			if ((echoes.voidorb != nil and echoes.voidorb.valid) or
					echoes.twirltimer > 0) and echoes.stuncooldown <= 0
				P_DoPlayerPain(target.player)
				PainSound(target)
				echoes.stuncooldown = TICRATE * 8
			elseif echoes.abyssactive > 0
				PainSound(target)
				echoes.stuncooldown = 0
				if echoes.abyss == nil or not echoes.abyss.valid
					echoes.abyss = ObjectSpawner(target, MT_ECHOES_ABYSS)
					echoes.abyss.protect = target.player
					echoes.abyss.tracer = target
					echoes.abyss.hoverheight = 0
					echoes.abyss.state = S_ABYSS_TELEPORT_IN
				else
					echoes.abyss.state = S_ABYSS_TELEPORT_OUT
				end
				if not echoes.abyss.dampause
					echoes.abyss.dampause = TICRATE / 3
					if echoes.abyss_shield != SH_NONE
						if echoes.abyss_shield == SH_ARMAGEDDON
							echoes.abyss.nuke = true
						end
						echoes.abyss_shield = SH_NONE
						target.player.powers[pw_shield] = SH_NONE
					end
					S_StartSound(echoes.abyss, sfx_abby2)
				end
			//otherwise if abyss is not active, echoes fucking dies
			else
				P_KillMobj(target, inflictor, source, damagetype)
			end
			return true
		end
	end
end

//ensure echoes doesn't take additional damage during damage state
local function DamageFailsafe(target, inflictor, source, damage, damagetype)
	if target != nil and target.valid and target.player != nil and
			target.player.valid and target.skin == "echoes" and
			((target.player.echoes.abyss != nil and
			target.player.echoes.abyss.valid and
			target.player.echoes.abyss.dampause != nil))
		if damagetype & DMG_DEATHMASK
			return
		end
		return false
	end
end

//playerthink function that handles the spawning and despawning of abyss
Echoes.AbyssSpawner = function(player)
	if not SkinCheck(player)
		return
	end
	
	local echoes = player.echoes
	//despawn abyss (if active) and clear variables if echoes is dead
	if player.playerstate == PST_DEAD or IsNiGHTS(player)
		if echoes.abyss != nil and echoes.abyss.valid and not 
				(echoes.abyss.dampause or echoes.abyss.detpause or 
				echoes.abyss.despawn)
			if IsNiGHTS(player)
				P_RemoveMobj(echoes.abyss)
			elseif echoes.abyssactive > 0
				echoes.abyss.detpause = TICRATE / 3
			else
				echoes.abyss.state = S_ABYSS_TELEPORT_OUT
				S_StartSound(player.mo, sfx_abby2)
				echoes.abyss.despawn = true
				echoes.abyss = nil
			end
		end
		echoes.abyssactive = 0
		echoes.abyss_shield = SH_NONE
		echoes.stuncooldown = 0
		if IsNiGHTS(player)
			player.mo.color = ColorOpposite(player.skincolor)
		elseif player.mo.color != player.skincolor
			player.mo.color = player.skincolor
		end
		return
	end
	
	if player.mo.color == ColorOpposite(player.skincolor)
		player.mo.color = player.skincolor
	end
	//prevent abyss from being active in ultimate mode
	if ultimatemode and echoes.abyssactive > 0
		echoes.abyssactive = 0
	end
	//tick down stun cooldown
	if echoes.stuncooldown > 0
		echoes.stuncooldown = $ - 1
	end
	//spawn abyss if he's supposed to be active
	if echoes.abyssactive > 0 and (echoes.abyss == nil or not 
			echoes.abyss.valid) and (echoes.voidorb == nil or not
			echoes.voidorb.valid) and echoes.twirltimer <= 0
			and not isSuper(player) and echoes.leapcharge <= 0
			and not echoes.shadowleap and (player.powers[pw_carry] != 19 or 
			not player.mo.boat) and not (echoes.hideabyss and 
			echoes.transform <= 0)
		local abyssx = P_ReturnThrustX(player.mo, player.drawangle + 
				ANGLE_90 + ANGLE_45, 60 * player.mo.scale)
		local abyssy = P_ReturnThrustY(player.mo, player.drawangle + 
				ANGLE_90 + ANGLE_45, 60 * player.mo.scale)
		echoes.abyss = ObjectSpawner(player.mo, MT_ECHOES_ABYSS, abyssx, 
				abyssy)
		echoes.abyss.protect = player
		echoes.abyss.tracer = player.mo
		echoes.abyss.state = S_ABYSS_TELEPORT_IN
		echoes.abyss.hoverheight = 0
		S_StartSound(echoes.abyss, sfx_abby3)
	//otherwise despawn him
	elseif echoes.abyssactive <= 0 and echoes.abyss != nil and 
			echoes.abyss.valid and not echoes.abyss.despawn
		echoes.abyss.state = S_ABYSS_TELEPORT_OUT
		S_StartSound(player.mo, sfx_abby2)
		echoes.abyss.despawn = true
		echoes.abyss = nil
	end
	//read and store shield data before removing to modify abyss' perks
	if player.powers[pw_shield] != SH_NONE and echoes.abyssactive > 0
			and not isSuper(player)
		if echoes.abyss != nil and echoes.abyss.valid and echoes.abyss.dampause
			player.powers[pw_shield] = SH_NONE
		else
			if player.powers[pw_shield] & SH_FORCE
				echoes.abyssactive = 2
			else
				echoes.abyssactive = 1
			end
			echoes.abyss_shield = player.powers[pw_shield]
			if echoes.abyss_shield & SH_PROTECTELECTRIC
				player.powers[pw_shield] = SH_PROTECTELECTRIC
			else
				player.powers[pw_shield] = SH_NONE
			end
		end
	elseif isSuper(player) and player.powers[pw_shield] != SH_NONE and
			echoes.abyss_shield == SH_NONE
		echoes.abyss_shield = player.powers[pw_shield]
	end
	//respawn abyss if echoes is vulnerable and hits a checkpoint
	if player.starpostnum != nil and player.starpostnum != echoes.lastcheckhit
		if echoes.abyssactive <= 0
			echoes.abyssactive = 1
		end
		echoes.lastcheckhit = player.starpostnum
	end
	//manage drowning
	local SequenceStuff = {
		[875] = true,
		[700] = true,
		[525] = true,
		[385] = true,
		[315] = true,
		[245] = true,
		[175] = true,
		[105] = true,
		[35] = true,
		[1] = true
	}
	if player.powers[pw_underwater] > 0
		if (echoes.abyssactive > 0 and G_CoopGametype()) or 
				echoes.abyss_shield & SH_PROTECTWATER
			//restore music if drowning theme had been playing
			if player.powers[pw_underwater] <= 11 * TICRATE
				player.powers[pw_underwater] = 30 * TICRATE
				P_RestoreMusic(player)
			end
			player.powers[pw_underwater] = 30 * TICRATE
		elseif not SequenceStuff[player.powers[pw_underwater] - 1] and
				leveltime % 2 == 0 and G_CoopGametype()
			player.powers[pw_underwater] = max($ - 1, 1)
		end
		if echoes.abyss_shield & SH_PROTECTELECTRIC
			echoes.abyss_shield = SH_NONE
		end
	end
	if player.powers[pw_spacetime] > 0
		if (echoes.abyssactive > 0 and G_CoopGametype()) or 
				echoes.abyss_shield & SH_PROTECTWATER
			player.powers[pw_spacetime] = 11 * TICRATE
		elseif not SequenceStuff[player.powers[pw_spacetime] - 1] and
				leveltime % 2 == 0 and G_CoopGametype()
			player.powers[pw_spacetime] = max($ - 1, 1)
		end
	end
end

//overwrite monitor pops to instead respawn abyss if echoes is vulnerable
function A_MonitorPop(actor, var1, var2)
	if actor.type != MT_EGGMAN_BOX and actor.target != nil and 
			actor.target.valid and actor.target.skin == "echoes" and
			actor.target.player != nil and actor.target.player.valid and
			actor.target.player.echoes != nil and 
			actor.target.player.echoes.abyssactive <= 0
		actor.target.player.echoes.abyssactive = 1
		actor.target = nil
	end
	super(actor, var1, var2)
end

//do similar thing for golden monitors
function A_GoldMonitorPop(actor, var1, var2)
	if actor.type != MT_EGGMAN_GOLDBOX and actor.target != nil and 
			actor.target.valid and actor.target.skin == "echoes" and
			actor.target.player != nil and actor.target.player.valid and
			actor.target.player.echoes != nil and 
			actor.target.player.echoes.abyssactive <= 0
		actor.target.player.echoes.abyssactive = 1
		actor.target = nil
	end
	super(actor, var1, var2)
end

//full thinker for Abyss
local function AbyssThinker(mo)
	//failsafe
	if mo.protect == nil or not mo.protect.valid or mo.protect.mo == nil or 
			not mo.protect.mo.valid or IsNiGHTS(mo.protect) or 
			mo.protect.mo.skin != "echoes" or ultimatemode
		P_RemoveMobj(mo)
		return
	end
	if (mo.protect.echoes.voidorb != nil and mo.protect.echoes.voidorb.valid) 
			or mo.protect.echoes.twirltimer > 0 or isSuper(mo.protect) or 
			mo.protect.echoes.leapcharge > 0 or mo.protect.echoes.shadowleap or 
			(mo.protect.echoes.hideabyss and mo.protect.echoes.transform <= 0)
			or (mo.protect.powers[pw_carry] == 19 or mo.protect.mo.boat)
		if not mo.despawn
			mo.state = S_ABYSS_TELEPORT_OUT
			if mo.protect != nil and mo.protect.valid
				S_StartSound(mo.protect.mo, sfx_abby2)
			end
			mo.despawn = true
		end
	end
	
	local protect = mo.protect
	if mo.tracer == nil or not mo.tracer.valid
		mo.tracer = mo.protect.mo
	end
	local tracer = mo.tracer
	mo.scale = tracer.scale
	mo.skin = tracer.skin
	mo.eflags = tracer.eflags
	ShieldColorFiller(mo, protect)
	//handle abyss' damage pause
	if mo.dampause != nil
		A_ForceStop(tracer, 0)
		protect.pflags = $ | PF_FULLSTASIS
		StateSetter(tracer, S_PLAY_PAIN)
		if mo.state == S_ABYSS_TELEPORT_OUT
			return
		end
		if mo.state == S_ABYSS_TELEPORT_IN and not mo.telesound
			S_StartSound(mo, sfx_abby3)
			mo.telesound = true
		elseif mo.state != S_ABYSS_TELEPORT_IN
			StateSetter(mo, S_ABYSS_PROTECT)
		end
		local abyssdist = 15
		mo.angle = protect.drawangle
		local abyssx = tracer.x + P_ReturnThrustX(mo, mo.angle, abyssdist *
				mo.scale)
		local abyssy = tracer.y + P_ReturnThrustY(mo, mo.angle, abyssdist *
				mo.scale)
		P_MoveOrigin(mo, abyssx, abyssy, ObjectZCoord(tracer, mo))
		if not mo.orb
			local bomb = ObjectSpawner(mo, MT_ECHOES_ORB)
			bomb.protectbomb = TICRATE / 2
			bomb.scale = $ / 2
			bomb.tracer = tracer
			bomb.nuke = mo.nuke
			mo.nuke = nil
			mo.orb = true
		end
		if mo.dampause > 0
			mo.dampause = $ - 1
		else
			local explode = ObjectSpawner(mo, MT_ECHOES_EFFECT)
			explode.state = S_ECHOES_RIPPLE
			explode.dispoffset = -1
			explode.color = ColorOpposite(protect.skincolor)
			explode.scale = $ * 3
			protect.rings = $ / 2
			P_DoPlayerPain(protect)
			protect.echoes.abyssactive = max($ - 1, 0)
			S_StartSound(tracer, sfx_abby1)
			mo.dampause = nil
			mo.orb = false
		end
		return
	end
	//handle abyss' death pause
	if mo.detpause != nil
		StateSetter(mo, S_ABYSS_DEAD)
		if not mo.orb
			local bomb = ObjectSpawner(mo, MT_ECHOES_ORB)
			bomb.protectbomb = TICRATE / 2
			bomb.scale = $ / 2
			bomb.tracer = tracer
			mo.orb = true
		end
		if mo.detpause > 0
			mo.detpause = $ - 1
		else
			local explode = ObjectSpawner(mo, MT_ECHOES_EFFECT)
			explode.state = S_ECHOES_RIPPLE
			explode.dispoffset = -1
			explode.color = ColorOpposite(protect.skincolor)
			explode.scale = $ * 3
			mo.state = S_ABYSS_TELEPORT_OUT
			S_StartSound(tracer, sfx_abby1)
			mo.despawn = true
			mo.detpause = nil
		end
		return
	end
	//have abyss despawn properly once the teleport animation concluded
	if mo.despawn
		if mo.state != S_ABYSS_TELEPORT_OUT
			P_RemoveMobj(mo)
		end
		return
	end
	//ensure no movement stuff runs while transformation sequence is active
	if protect.echoes.abyssactive == 3
		return
	end
	//change abyss' state and position depending on movement
	if mo.state != S_ABYSS_TELEPORT_IN and mo.state != S_ABYSS_TELEPORT_OUT
		if SpeedCalc(protect) >= 20 * mo.scale
			StateSetter(mo, S_ABYSS_RUN)
		else
			StateSetter(mo, S_ABYSS_IDLE)
		end
	end
	mo.angle = protect.drawangle
	local abyssx, abyssy
	local destx = tracer.x + P_ReturnThrustX(mo, mo.angle + ANGLE_90 + 
			ANGLE_45, 60 * mo.scale)
	local desty = tracer.y + P_ReturnThrustY(mo, mo.angle + ANGLE_90 + 
			ANGLE_45, 60 * mo.scale)
	local distfrom = R_PointToDist2(mo.x, mo.y, destx, desty)
	local abyssx = destx
	local abyssy = desty
	if distfrom > 0 and not protect.echoes.outoftouch
		local pangle = R_PointToAngle2(mo.x, mo.y, destx, desty) + ANGLE_180
		local dist = min(distfrom - distfrom / 5, 100 * mo.scale)
		abyssx = $ + P_ReturnThrustX(mo, pangle, dist)
		abyssy = $ + P_ReturnThrustY(mo, pangle, dist)
	else
		abyssx = mo.x
		abyssy = mo.y
	end
	if leveltime % 5 == 0
		if not mo.descend
			mo.hoverheight = $ + mo.scale / 2
			if mo.hoverheight >= mo.scale + mo.scale / 2
				mo.descend = true
			end
		else
			mo.hoverheight = $ - mo.scale / 2
			if mo.hoverheight <= (mo.scale + mo.scale / 2) * -1
				mo.descend = false
			end
		end
	end
	local zlagoff = min(tracer.momz * -1 * 3, 50 * mo.scale)
	zlagoff = max($, -50 * mo.scale)
	local abyssz = mo.hoverheight + zlagoff + ObjectZCoord(tracer, mo, 
			mo.scale * 3)
	if distfrom > 200 * mo.scale
		P_SetOrigin(mo, abyssx, abyssy, abyssz)
	elseif P_TryMove(mo, abyssx, abyssy, true)
		P_MoveOrigin(mo, abyssx, abyssy, abyssz)
	end
	//spawn particles and change appearance depending on shield type
	if protect.echoes.abyss_shield != SH_NONE or protect.echoes.abyssactive > 1
		ShieldColorFiller(ParticleSpawner(mo), protect)
		mo.blendmode = AST_SUBTRACT
	elseif mo.blendmode == AST_SUBTRACT
		mo.blendmode = AST_TRANSLUCENT
	end
	//kinda annoyed i have to do this just for collecting rings but whatevs
	local ringcand = SearchRadius(mo, tracer, mo.height + mo.scale * 5)
	local therering = false
	for i, rin in ipairs(ringcand) do
		if isRing(rin)
			therering = true
			break
		end
	end
	if therering
		mo.flags = $ & ~MF_NOCLIP
	elseif not (mo.flags & MF_NOCLIP)
		mo.flags = $ | MF_NOCLIP
	end
end

//have abyss collect rings and blue spheres
local function AbyssRings(abyss, ring)
	if abyss != nil and abyss.valid and abyss.protect != nil and 
			abyss.protect.valid and ring != nil and ring.valid and isRing(ring)
		if (ring.z + ring.height >= abyss.z and ring.z + ring.height <= 
				abyss.z + abyss.height) or (ring.z >= abyss.z and
				ring.z <= abyss.z + abyss.height)
			P_GivePlayerRings(abyss.protect, 1)
			if ring.type == MT_FLINGCOIN or ring.type == MT_COIN
				S_StartSound(abyss, sfx_mario4)
			else
				S_StartSound(abyss, sfx_itemup)
			end
			P_KillMobj(ring, abyss.tracer)
		end
	else
		return false
	end
end

//hooks
addHook("MobjDamage", FragilityDamage, MT_PLAYER)
addHook("ShouldDamage", DamageFailsafe, MT_PLAYER)
addHook("MobjThinker", AbyssThinker, MT_ECHOES_ABYSS)
addHook("MobjCollide", AbyssRings, MT_ECHOES_ABYSS)
addHook("MobjMoveCollide", AbyssRings, MT_ECHOES_ABYSS)
//have a quick hook to have abyss phase through walls
//because noclip objects don't trigger mobjcollide hooks even with MF_SOLID
//thanks 2.2.15 >:(
addHook("MobjLineCollide", function(mo, line)
	return false
end, MT_ECHOES_ABYSS)
//this makes it so echoes can't grab air bubbles while not vulnerable
addHook("TouchSpecial", function(special, toucher)
	if toucher != nil and toucher.valid and toucher.skin == "echoes" and 
			toucher.player != nil and toucher.player.valid and 
			toucher.player.echoes != nil and G_CoopGametype() and
			toucher.player.echoes.abyssactive > 0
		return true
	end
end, MT_EXTRALARGEBUBBLE)