//Echoes' jump abilities, coded by Emmette Clyde w/ slight mods by Fav

local SkinCheck = Echoes.SkinCheck
local ObjectSpawner = Echoes.ObjectSpawner
local FailVibeCheck = Echoes.FailVibeCheck
local StateSetter = Echoes.StateSetter
local JumpManager = Echoes.JumpManager
local SearchRadius = Echoes.SearchRadius
local FoxOnGround = Echoes.FoxOnGround
local SpeedCalc = Echoes.SpeedCalc
local ShieldColorFiller = Echoes.ShieldColorFiller
local StateReset = Echoes.StateReset
local ObjectZCoord = Echoes.ObjectZCoord
local isRing = Echoes.isRing
local isSuper = Echoes.isSuper
local GetFloorZ = Echoes.GetFloorZ
local ParticleSpawner = Echoes.ParticleSpawner
local CamShake = Echoes.CamShake
local PainSound = Echoes.PainSound
local IsNerfed = Echoes.IsNerfed
local PlayerCam = Echoes.PlayerCam

//set up trigger for void glide
local function VoidGlide(player)
	if not SkinCheck(player)
		return
	end
	
	local echoes = player.echoes
	if not echoes.jumppress
		echoes.jumppress = true
	end
	
	if (echoes.phanttwirl > 1 and echoes.twirltimer > 0) or echoes.shadowleap
			or echoes.transform >= 0 or FailVibeCheck(player) or 
			echoes.voidtumble
		return
	end
	if not (player.pflags & PF_THOKKED)
		player.pflags = $ | PF_THOKKED & ~PF_STARTJUMP & ~PF_SPINNING
		echoes.voidlaunch = true
		if echoes.twirltimer > 0
			echoes.twirltimer = 0
		end
	end
	if (echoes.voidorb == nil or not echoes.voidorb.valid) and 
			echoes.voidstamina > 0 and echoes.voidcooldown <= 0
		echoes.voidorb = ObjectSpawner(player.mo, MT_ECHOES_ORB)
		echoes.voidorb.tracer = player.mo
		if echoes.abyssactive <= 0
			S_StartSound(player.mo, sfx_vogl2)
		else
			S_StartSound(player.mo, sfx_vogl1)
		end
	end
	return true
end

//function to make removing the orb easier
local function OrbRemover(mo)
	if mo != nil and mo.valid
		mo.removing = true
		if mo.pullers != nil
			for i, p in ipairs(mo.pullers) do
				if p != nil and p.valid
					if p.tracer != nil and p.tracer.valid and p.tracer.health 
							and p.tracer.state != p.tracer.info.deathstate
						p.tracer.flags = $ & ~MF_NOCLIP & ~MF_NOCLIPHEIGHT
						if p.tracer.flags & MF_ENEMY
							P_DamageMobj(p.tracer, mo, mo.tracer, 1)
						elseif isRing(p.tracer)
							P_GivePlayerRings(mo.tracer.player, 1)
							if p.tracer.type == MT_FLINGCOIN or 
									p.tracer.type == MT_COIN
								S_StartSound(mo.tracer, sfx_mario4)
							else
								S_StartSound(mo.tracer, sfx_itemup)
							end
							P_KillMobj(p.tracer, mo.tracer)
						else
							P_KillMobj(p.tracer, mo, mo.tracer)
						end
					end
					P_RemoveMobj(p)
				end
				table.remove(mo.pullers, i)
			end
		end
		if mo.pullers == nil or unpack(mo.pullers) == nil
			P_RemoveMobj(mo)
		end
	end
end

//actual thinker for void glide
Echoes.VoidGlideThinker = function(player)
	if not SkinCheck(player)
		return
	end
	
	local echoes = player.echoes
	if not (player.cmd.buttons & BT_JUMP) and echoes.jumppress
		echoes.jumppress = false
	end
	if FailVibeCheck(player) or not (player.pflags & PF_THOKKED) or
			echoes.transform >= 0
		if echoes.voidtumble
			if FoxOnGround(player.mo)
				JumpManager(player.mo, 5 * FRACUNIT)
				StateSetter(player.mo, S_PLAY_FALL)
			end
			echoes.voidtumble = false
			player.mo.spriteroll = 0
			player.powers[pw_strong] = STR_NONE
		end
		if echoes.voidorb != nil and echoes.voidorb.valid
			S_StartSound(player.mo, sfx_vogl7)
			OrbRemover(echoes.voidorb)
			echoes.voidorb = nil
		end
		echoes.voidheight = 0
		echoes.voidlaunch = false
		echoes.voidstamina = TICRATE * 5 + TICRATE / 2
		echoes.voidcooldown = -1
		StateReset(player.mo, {S_PLAY_FLY, S_PLAY_SWIM, S_ECHOES_TIRED})
		return
	end
	
	//handle the initial launch/takeoff
	local voidflags = STR_ATTACK | STR_WALL | STR_FLOOR | STR_CEILING | 
			STR_SPIKE
	local glidestate
	if player.mo.eflags & MFE_UNDERWATER or player.mo.eflags & MFE_GOOWATER
		glidestate = S_PLAY_SWIM
	else
		glidestate = S_PLAY_FLY
	end
	if echoes.voidlaunch
		StateSetter(player.mo, glidestate, voidflags)
		JumpManager(player.mo, echoes.voidheight * FRACUNIT)
		local ghost = P_SpawnGhostMobj(player.mo)
		if echoes.abyssactive > 0
			ghost.blendmode = AST_MODULATE
		end
		echoes.voidheight = $ + 1
		local void_height_max = 15
		if echoes.abyssactive <= 0 or IsNerfed()
			void_height_max = 9
		end
		if echoes.voidheight >= void_height_max or not echoes.jumppress
			echoes.voidlaunch = false
			if not echoes.jumppress and echoes.voidheight > 9
				echoes.voidheight = 9
			end
		end
	//handle the glide that follows
	else
		if echoes.voidheight > 9 and not echoes.jumppress
			echoes.voidheight = 9
		end
		if echoes.voidorb != nil and echoes.voidorb.valid
			if echoes.voidstamina <= TICRATE * 2
				StateSetter(player.mo, S_ECHOES_TIRED, voidflags)
				local cam = PlayerCam(player)
				if echoes.sweat == nil or not echoes.sweat.valid
					echoes.sweat = ObjectSpawner(player.mo, MT_ECHOES_EFFECT)
					echoes.sweat.state = S_ECHOES_SWEAT
					echoes.sweat.dispoffset = 3
					echoes.sweat.scale = $ + ($ / 5)
					S_StartSound(player.mo, sfx_s1b6)
				end
				local pang = cam.angle - ANGLE_90
				local x = player.mo.x + P_ReturnThrustX(player.mo, pang, 
						25 * player.mo.scale)
				local y = player.mo.y + P_ReturnThrustY(player.mo, pang, 
						25 * player.mo.scale)
				local z = ObjectZCoord(player.mo, echoes.sweat, player.mo.height)
				P_MoveOrigin(echoes.sweat, x, y, z)
				echoes.sweat.angle = pang
			else
				StateSetter(player.mo, glidestate, voidflags)
			end
			if leveltime % 3 == 0
				echoes.voidheight = max($ - 1, -1)
			end
			if (echoes.jumppress or echoes.voidheight > -1) and
					echoes.voidcooldown <= 0
				JumpManager(player.mo, echoes.voidheight * FRACUNIT)
			end
			if not isSuper(player)
				echoes.voidstamina = max($ - 1, 0)
			end
			if (echoes.abyssactive <= 0 or IsNerfed()) and leveltime % 2 == 0
				echoes.voidstamina = max($ - 1, 0)
			end
		else
			if echoes.voidstamina <= 0 and not echoes.voidtumble
				echoes.voidtumble = true
				PainSound(player.mo)
			end
			if echoes.voidtumble
				StateSetter(player.mo, S_PLAY_STUN, STR_ATTACK)
				player.pflags = $ | PF_FULLSTASIS
				if leveltime % 2 == 0
					player.mo.spriteroll = $ + ANGLE_45
				end
				player.drawangle = $ + ANGLE_22h
				if player.mo.eflags & MFE_GOOWATER
					echoes.voidtumble = false
					player.mo.spriteroll = 0
					player.powers[pw_strong] = STR_NONE
				end
			else
				if echoes.voidcooldown > -1
					echoes.voidcooldown = max($ - 1, 0)
				end
				echoes.voidheight = 0
				StateReset(player.mo, {S_PLAY_FLY, S_PLAY_SWIM, S_ECHOES_TIRED})
			end
		end
	end
end

//have the orb bust breakable floors and walls
local function BustableCheck(mo, sec)
	for fof in sec.ffloors() do
		if fof.flags & FF_EXISTS and fof.flags & FF_BUSTUP and
				((mo.z + mo.height >= fof.bottomheight and 
				mo.z - mo.height <= fof.topheight) or 
				(fof.bottomheight >= mo.z - mo.height and
				fof.topheight <= mo.z + mo.height) or 
				(mo.z - mo.height <= fof.topheight and
				mo.z - mo.height >= fof.bottomheight) or
				(mo.z + mo.height <= fof.topheight and
				mo.z + mo.height >= fof.bottomheight) or
				(ObjectZCoord(mo) >= fof.bottomheight and 
				ObjectZCoord(mo)<= fof.topheight))
			EV_CrumbleChain(fof)
		end
	end
end

//thinker for the orb since using an action won't work quickly enough
local function OrbThinker(mo)
	if mo.removing
		mo.state = S_INVISIBLE
		OrbRemover(mo)
		return
	end
	//behavior if spawned by abyss protecting echoes
	if mo.protectbomb != nil
		mo.blendmode = AST_SUBTRACT
		if mo.nuke and not mo.colorized
			mo.colorized = true
			mo.color = SKINCOLOR_VOLCANIC
		elseif not mo.nuke and mo.tracer != nil and mo.tracer.valid and
				mo.tracer.player != nil and mo.tracer.player.valid
			mo.colorized = false
			mo.color = ColorOpposite(mo.tracer.player.skincolor)
		end
		//destroy enemies and hazards within the vicinity
		local orbvictims = SearchRadius(mo, mo.tracer, mo.radius)
		for i, v in ipairs(orbvictims) do
			if v != nil and v.valid and v.health and v.state != 
					v.info.deathstate and not (v.flags & MF_MONITOR
					or isRing(v))
				if v.flags & MF_ENEMY or v.flags & MF_BOSS
					P_DamageMobj(v, mo, mo.tracer, 1)
				else
					P_KillMobj(v, mo, mo.tracer)
				end
			end
		end
		//grow/shrink the orb based on timer
		if mo.protectbomb > 0
			if mo.protectbomb <= TICRATE / 6
				mo.scale = $ - mo.tracer.scale
			else
				mo.scale = $ + mo.tracer.scale / 8
				if mo.nuke
					mo.scale = $ + mo.tracer.scale / 8
				end
			end
			mo.protectbomb = $ - 1
		else
			P_RemoveMobj(mo)
		end
	//otherwise assume behavior for void glide
	elseif mo.tracer != nil and mo.tracer.valid and mo.tracer.skin == "echoes"
		//failsafe
		if mo.tracer.player != nil and mo.tracer.player.valid and (not 
				(mo.tracer.player.pflags & PF_THOKKED) or 
				FailVibeCheck(mo.tracer.player) or mo.tracer.player.echoes == 
				nil or mo.tracer.player.echoes.voidorb == nil or not
				mo.tracer.player.echoes.voidorb.valid or
				mo.tracer.player.echoes.voidorb != mo) or 
				mo.tracer.player.echoes.transform >= 0 or
				mo.tracer.player.echoes.voidtumble
			OrbRemover(mo)
			return
		end
		mo.eflags = mo.tracer.eflags
		local z = ObjectZCoord(mo.tracer, mo, mo.tracer.height / 2)
		P_MoveOrigin(mo, mo.tracer.x, mo.tracer.y, z)
		if mo.voidswirl == nil or not mo.voidswirl.valid
			mo.voidswirl = ObjectSpawner(mo, MT_ECHOES_SWIRL)
			mo.voidswirl.tracer = mo
		end
		local echoes = mo.tracer.player.echoes
		local void_launch_max = mo.tracer.scale * 2
		local void_glide_size = mo.tracer.scale - mo.tracer.scale / 5
		if isSuper(mo.tracer.player)
			void_glide_size = mo.tracer.scale + mo.tracer.scale / 5
		elseif echoes.abyssactive <= 0 or IsNerfed()
			void_launch_max = mo.tracer.scale + mo.tracer.scale / 2
			void_glide_size = mo.tracer.scale - mo.tracer.scale / 3
		end
		//pull in victims towards echoes
		local maxpull = mo.radius / mo.scale * void_launch_max
		local pullradius = min(mo.radius * 2, maxpull + maxpull / 2)
		local pullvictims = SearchRadius(mo, mo.tracer, pullradius)
		if mo.pullers == nil
			mo.pullers = {}
		end
		for i, v in ipairs(pullvictims) do
			if v != nil and v.valid and v.health and v.state != 
					v.info.deathstate and v.type != MT_SPIKE and
					v.type != MT_WALLSPIKE and not (v.flags & MF_BOSS or
					v.flags & MF_MISSILE or v.flags & MF_MONITOR or v.flags2 & 
					MF2_DONTDRAW) and v.sprite != SPR_NULL
				local intable = false
				for i, p in ipairs(mo.pullers) do
					if p != nil and p.valid and p.tracer != nil and
							p.tracer.valid and p.tracer == v
						intable = true
					end
				end
				if not intable
					local puller = ObjectSpawner(v, MT_ECHOES_DUMMY)
					puller.tracer = v
					puller.target = mo.tracer
					local h_dist = R_PointToDist2(mo.tracer.x, mo.tracer.y, 
							v.x, v.y)
					local temp = R_PointToDist2(0, mo.tracer.z, h_dist, v.z)
					puller.pullrad = min(temp, pullradius)
					puller.angle = R_PointToAngle2(mo.tracer.x, mo.tracer.y, 
							v.x, v.y)
					table.insert(mo.pullers, puller)
				end
			end
		end
		//have pulled victims swirl around echoes as they get pulled in
		for i, p in ipairs(mo.pullers) do
			if p != nil and p.valid
				if p.tracer != nil and p.tracer.valid and p.tracer.health
						and p.tracer.state != p.tracer.info.deathstate
						and p.target != nil and p.target.valid
					//var1(placement) = (vertical offset << 16) + radius
					//var2(rotation) = (horizontal speed << 16) + vertical speed
					local var1 = (0 << 16) + p.pullrad / mo.tracer.scale
					local var2 = (150 << 16) + 0
					A_Custom3DRotate(p, var1, var2)
					//give the actual objects noclip just in case
					p.tracer.flags = $ | MF_NOCLIP | MF_NOCLIPHEIGHT
					P_MoveOrigin(p.tracer, p.x, p.y, p.z)
					p.pullrad = max($ - mo.tracer.scale * 10, 0)
					p.pullrad = min($, mo.radius + mo.radius / 2)
					if p.pullrad <= 0
						p.tracer.flags = $ & ~MF_NOCLIP & ~MF_NOCLIPHEIGHT
						P_RemoveMobj(p)
						table.remove(mo.pullers, i)
					end
				else
					if p.tracer != nil and p.tracer.valid and p.tracer.health
							and p.tracer.state != p.tracer.info.deathstate
						p.tracer.flags = $ & ~MF_NOCLIP & ~MF_NOCLIPHEIGHT
					end
					P_RemoveMobj(p)
					table.remove(mo.pullers, i)
				end
			else
				table.remove(mo.pullers, i)
			end
		end
		//otherwise damage those that are close enough to echoes
		local damradius = mo.tracer.radius + mo.tracer.radius / 2
		local orbvictims = SearchRadius(mo, mo.tracer, damradius)
		for i, v in ipairs(orbvictims) do
			if v != nil and v.valid and v.health and v.state != 
					v.info.deathstate and not isRing(v)
				if v.flags & MF_ENEMY or v.flags & MF_BOSS or v.flags & 
						MF_MONITOR
					P_DamageMobj(v, mo, mo.tracer, 1)
				else
					P_KillMobj(v, mo, mo.tracer)
				end
			end
		end
		//separate search for spikes & monitors since they need to be damaged by contact
		local spikevictims = SearchRadius(mo, mo.tracer, mo.radius)
		for i, s in ipairs(spikevictims) do
			if s != nil and s.valid and s.health and s.state != 
					s.info.deathstate and (s.type == MT_SPIKE or 
					s.type == MT_WALLSPIKE or s.flags & MF_MONITOR or
					s.flags & MF_BOSS)
				if s.flags & MF_MONITOR or s.flags & MF_BOSS
					P_DamageMobj(s, mo, mo.tracer, 1)
				else
					P_KillMobj(s, mo, mo.tracer)
				end
			end
		end
		//scan for lines we're colliding with so that we can bust bustables
		//thank you to Lach for letting me reference Aether's code for this
		searchBlockmap("lines", function(mo, fline)
			if fline != nil and fline.valid
				local x, y = P_ClosestPointOnLine(mo.x, mo.y, fline)
				local v1 = fline.v1
				local v2 = fline.v2
				local minx = min(v1.x, v2.x)
				local maxx = max(v1.x, v2.x)
				local miny = min(v1.y, v2.y)
				local maxy = max(v1.y, v2.y)
				x = min(max($, minx), maxx)
				y = min(max($, miny), maxy)
				local dist = R_PointToDist2(mo.x, mo.y, x, y)
				if dist <= mo.radius
					for i, sec in ipairs({fline.frontsector, 
							fline.backsector}) do
						BustableCheck(mo, sec)
					end
				end
			end
		end, mo, mo.x - mo.radius, mo.x + mo.radius, mo.y - mo.radius, 
				mo.y + mo.radius)
		//sound effects
		if echoes.voidstamina == TICRATE * 2
			S_StopSound(mo)
			mo.soundcue = 0
		end
		if mo.soundcue == nil or mo.soundcue <= 0
			if echoes.voidstamina <= TICRATE * 2
				S_StartSound(mo, sfx_vogl4)
			else
				S_StartSound(mo, sfx_vogl3)
			end
			mo.soundcue = TICRATE * 2
		elseif mo.soundcue > 0
			mo.soundcue = $ - 1
		end
		//change size and appearance based on situation
		if echoes.abyssactive <= 0
			StateSetter(mo, S_ECHOES_ORB_WEAK)
		else
			StateSetter(mo, S_ECHOES_ORB)
			mo.blendmode = AST_SUBTRACT
		end
		ShieldColorFiller(mo, mo.tracer.player)
		if mo.rippletime == nil or mo.rippletime <= 0
			mo.ripple = ObjectSpawner(mo, MT_ECHOES_EFFECT)
			mo.ripple.state = S_ECHOES_RIPPLE
			mo.ripple.dispoffset = -1
			ShieldColorFiller(mo.ripple, mo.tracer.player)
			mo.rippletime = TICRATE
		elseif mo.rippletime > 0
			mo.rippletime = $ - 1
		end
		if mo.ripple != nil and mo.ripple.valid
			mo.ripple.scale = min(mo.scale * 3, void_launch_max + 
					void_launch_max / 2)
			P_MoveOrigin(mo.ripple, mo.x, mo.y, ObjectZCoord(mo))
		end
		if echoes.voidlaunch
			mo.scale = min($ + mo.tracer.scale / 13, void_launch_max)
		else
			if echoes.jumppress and not mo.fastclose
				if mo.scale > void_glide_size
					mo.scale = max($ - mo.tracer.scale / 60, void_glide_size)
				else
					mo.scale = min($ + mo.tracer.scale / 20, void_glide_size)
				end
				if echoes.voidstamina <= 0
					S_StartSound(mo.tracer, sfx_vogl7)
					OrbRemover(mo)
				end
			else
				mo.scale = max($ - mo.tracer.scale / 40, 0)
				if mo.fastclose
					mo.scale = max($ - mo.tracer.scale / 8, 0)
				end
				if mo.scale < mo.tracer.scale / 15
					S_StartSound(mo.tracer, sfx_vogl6)
					OrbRemover(mo)
				end
			end
		end
	//failsafe
	else
		OrbRemover(mo)
	end
end

//thinker for the swirl
local function SwirlThinker(mo)
	mo.blendmode = AST_SUBTRACT
	if mo.tracer != nil and mo.tracer.valid
		mo.scale = mo.tracer.scale
		mo.eflags = mo.tracer.eflags
		ShieldColorFiller(mo, mo.tracer.tracer.player)
		P_MoveOrigin(mo, mo.tracer.x, mo.tracer.y, mo.tracer.z)
	else
		P_RemoveMobj(mo)
	end
end

//have echoes bounce off stuff she collides with while void gliding
Echoes.VoidBounce = function(mo, thing, line)
	if mo != nil and mo.valid and mo.player != nil and mo.player.valid
			and mo.skin == "echoes"
		if mo.player.echoes != nil and ((mo.player.echoes.voidorb != nil
				and mo.player.echoes.voidorb.valid) or 
				mo.player.echoes.voidtumble)
			local bsped = max(SpeedCalc(mo.player) / 2, 5 * mo.scale)
			local BonkSound = {
				sfx_bonk1, sfx_bonk2, sfx_bonk3
			}
			if line != nil and line.valid
				if mo.flags2 & MF2_TWOD
					P_InstaThrust(mo, mo.player.drawangle + ANGLE_180, bsped)
				else
					P_BounceMove(mo)
				end
				if mo.player.echoes.voidtumble
					S_StartSound(mo, BonkSound[P_RandomRange(1, 3)])
				else
					S_StartSound(mo, sfx_vogl5)
				end
				CamShake(mo.player)
			elseif thing != nil and thing.valid
				local intable = false
				if mo.player.echoes.voidorb != nil and 
						mo.player.echoes.voidorb.valid and 
						mo.player.echoes.voidorb.pullers != nil
					for i, p in ipairs(mo.player.echoes.voidorb.pullers) do
						if p != nil and p.valid and p.tracer != nil and
								p.tracer.valid and p.tracer == thing
							intable = true
						end
					end
				end
				if not intable
					local bangle = R_PointToAngle2(thing.x, thing.y, mo.x, mo.y)
					if bangle == nil or mo.flags2 & MF2_TWOD
						bangle = mo.player.drawangle + ANGLE_180
					end
					P_Thrust(mo, bangle, bsped)
					if mo.player.echoes.voidtumble
						S_StartSound(mo, BonkSound[P_RandomRange(1, 3)])
					else
						S_StartSound(mo, sfx_vogl5)
					end
					CamShake(mo.player)
				end
			end
		end
	end
end

//charge up shadow leap
local function LeapCharge(player)
	if not SkinCheck(player) or (player.echoes != nil and 
			(player.echoes.transform >= 0 or player.echoes.voidtumble))
			or FailVibeCheck(player)
		return
	end
	
	local max_leap_charge = Echoes.max_leap_charge
	local echoes = player.echoes
	if echoes.crouch
		S_StartSound(player.mo, sfx_shlp1)
		echoes.leapcharge = $ + 1
		player.height = skins[player.mo.skin].height
		local ripple = ObjectSpawner(player.mo, MT_ECHOES_EFFECT)
		ripple.state = S_ECHOES_RIPPLE
		ripple.frame = $ | FF_FLOORSPRITE
		ShieldColorFiller(ripple, player)
		if isSuper(player)
			ripple.scale = $ * 2
		end
		P_SetOrigin(ripple, player.mo.x, player.mo.y, GetFloorZ(player.mo))
		echoes.chargeripple = 20
		echoes.crouch = false
		return true
	elseif echoes.leapcharge > 0
		StateSetter(player.mo, S_ECHOES_LEAPCHARGE)
		if FixedHypot(player.mo.momx, player.mo.momy) >= 2*player.mo.scale
			player.powers[pw_nocontrol] = 2 + 32768
			P_Thrust(player.mo, R_PointToAngle2(player.mo.x+player.mo.momx, 
					player.mo.y+player.mo.momy, player.mo.x, player.mo.y), 
					(player.mo.scale*2)/3)
		else
			player.mo.momx = 0 + player.cmomx
			player.mo.momy = 0 + player.cmomy
		end
		local particle = ParticleSpawner(player.mo)
		if echoes.leapcharge >= max_leap_charge
			particle.sprite = SPR_PTSP
		end
		ShieldColorFiller(particle, player)
		if echoes.leapcharge > 20
			if leveltime % 5 == 0
				if not echoes.chargerise
					echoes.chargeheight = $ + FRACUNIT / 2
					if echoes.chargeheight >= FRACUNIT + FRACUNIT / 2
						echoes.chargerise = true
					end
				else
					echoes.chargeheight = $ - FRACUNIT / 2
					if echoes.chargeheight <= (FRACUNIT + FRACUNIT / 2) * -1
						echoes.chargerise = false
					end
				end
			end
		else
			echoes.chargeheight = FRACUNIT
		end
		JumpManager(player.mo, echoes.chargeheight)
		echoes.leapcharge = min($ + 1, max_leap_charge)
		if isSuper(player)
			echoes.leapcharge = min($ + 1, max_leap_charge)
		end
		if echoes.leapcharge >= max_leap_charge and not echoes.maxleapchrg
			S_StartSound(player.mo, sfx_shlp2)
			echoes.maxleapchrg = true
		end
		if echoes.chargeripple <= 0
			local ripple = ObjectSpawner(player.mo, MT_ECHOES_EFFECT)
			ripple.state = S_ECHOES_RIPPLE
			ripple.frame = $ | FF_FLOORSPRITE
			ShieldColorFiller(ripple, player)
			if isSuper(player)
				ripple.scale = $ * 2
			end
			P_SetOrigin(ripple, player.mo.x, player.mo.y, GetFloorZ(player.mo))
			echoes.chargeripple = 20
		else
			echoes.chargeripple = $ - 1
		end
		if leveltime % 5 == 0
			CamShake(player)
		end
		return true
	end
end

//hooks
addHook("AbilitySpecial", VoidGlide)
addHook("MobjThinker", OrbThinker, MT_ECHOES_ORB)
addHook("MobjThinker", SwirlThinker, MT_ECHOES_SWIRL)
addHook("JumpSpecial", LeapCharge)