//Echoes' spin abilities, coded by Emmette Clyde

local SkinCheck = Echoes.SkinCheck
local FailVibeCheck = Echoes.FailVibeCheck
local FoxOnGround = Echoes.FoxOnGround
local JumpManager = Echoes.JumpManager
local StateSetter = Echoes.StateSetter
local StateReset = Echoes.StateReset
local ObjectSpawner = Echoes.ObjectSpawner
local SearchRadius = Echoes.SearchRadius
local ParticleSpawner = Echoes.ParticleSpawner
local isFalling = Echoes.isFalling
local ObjectZCoord = Echoes.ObjectZCoord
local ShieldColorFiller = Echoes.ShieldColorFiller
local isRing = Echoes.isRing
local isSuper = Echoes.isSuper
local CamShake = Echoes.CamShake
local PainSound = Echoes.PainSound
local IsNerfed = Echoes.IsNerfed

local phant_twirl_maxtime = TICRATE

//set up functions and actions for the swirling fire effects
local function OrbitRemover(mo)
	if mo != nil and mo.valid
		mo.removing = true
		if mo.flames != nil
			for i, fl in ipairs(mo.flames) do
				if fl != nil and fl.valid
					P_RemoveMobj(fl)
				end
				table.remove(mo.flames, i)
			end
		end
		if mo.flames2 != nil
			for i, fl in ipairs(mo.flames2) do
				if fl != nil and fl.valid
					P_RemoveMobj(fl)
				end
				table.remove(mo.flames2, i)
			end
		end
		if (mo.flames == nil or unpack(mo.flames) == nil) and (mo.flames2 == nil
				or unpack(mo.flames2) == nil)
			P_RemoveMobj(mo)
		end
	end
end

function A_EchoesFlameOrbit(actor, var1, var2)
	//failsafes
	if actor == nil or not actor.valid
		return
	end
	if actor.removing
		if actor.flames != nil
			for i, fl in ipairs(actor.flames)
				if fl != nil and fl.valid
					fl.state = S_INVISIBLE
				end
			end
		end
		if actor.flames2 != nil
			for i, fl in ipairs(actor.flames2)
				if fl != nil and fl.valid
					fl.state = S_INVISIBLE
				end
			end
		end
		OrbitRemover(actor)
		return
	end
	local tracer = actor.tracer
	if tracer == nil or not tracer.valid or tracer.player == nil or not
			tracer.player.valid or tracer.skin != "echoes" or 
			tracer.player.echoes == nil
		OrbitRemover(actor)
		return
	end
	local echoes = tracer.player.echoes
	//control effects for the stomp landing
	if actor.stomp
		if actor.flames == nil
			actor.flames = {}
			for i = 0, 360, 45 do
				local flame = ObjectSpawner(actor, MT_ECHOES_EFFECT)
				flame.state = S_ECHOES_TWIRLFLAME
				flame.spriteyscale = $ / 2
				flame.target = actor
				flame.angle = FixedAngle(i * FRACUNIT)
				ShieldColorFiller(flame, tracer.player)
				flame.fade = 0
				table.insert(actor.flames, flame)
			end
		end
		if actor.xplradius == nil
			actor.xplradius = actor.radius
		end
		for i, fl in ipairs(actor.flames) do
			if fl != nil and fl.valid and fl.fade < 10
				local var1 = (0 << 16) + actor.xplradius / actor.scale
				local var2 = (150 << 16) + 0
				A_Custom3DRotate(fl, var1, var2)
				fl.blendmode = AST_SUBTRACT
				fl.frame = $ | (fl.fade << 16)
				local incmagn = max(actor.minrad / actor.scale / 250, 1)
				actor.xplradius = $ + FRACUNIT * incmagn
				if actor.xplradius >= actor.minrad / 5
					fl.fade = $ + 1
				end
				ShieldColorFiller(fl, tracer.player)
			else
				if fl != nil and fl.valid
					P_RemoveMobj(fl)
				end
				table.remove(actor.flames, i)
			end
		end
		if unpack(actor.flames) == nil
			P_RemoveMobj(actor)
		end
	//otherwise swirl around echoes while phantom twirling
	elseif echoes.twirltimer > 0 and not FailVibeCheck(tracer.player)
		P_MoveOrigin(actor, tracer.x, tracer.y, ObjectZCoord(tracer, actor, 
				tracer.height / 2))
		if actor.flames == nil
			actor.flames = {}
			for i = 0, 360, 90 do
				local flame = ObjectSpawner(actor, MT_ECHOES_EFFECT)
				flame.state = S_ECHOES_TWIRLFLAME
				flame.spriteyscale = $ / 2
				if isSuper(tracer.player)
					flame.scale = $ + $ / 2
				end
				flame.target = actor
				flame.angle = FixedAngle(i * FRACUNIT)
				ShieldColorFiller(flame, tracer.player)
				if echoes.abyssactive > 0
					flame.fade = 0
				else
					flame.fade = 5
				end
				table.insert(actor.flames, flame)
			end
		end
		if isSuper(tracer.player) and actor.flames2 == nil
			actor.flames2 = {}
			for i = 0, 360, 45 do
				local flame = ObjectSpawner(actor, MT_ECHOES_EFFECT)
				flame.state = S_ECHOES_TWIRLFLAME
				flame.spriteyscale = $ / 2
				flame.target = actor
				ShieldColorFiller(flame, tracer.player)
				flame.angle = FixedAngle(i * FRACUNIT)
				flame.fade = 0
				table.insert(actor.flames2, flame)
			end
		end
		for i, fl in ipairs(actor.flames) do
			if fl != nil and fl.valid and fl.fade < 10
				local var1 = (0 << 16) + tracer.radius * 2 / fl.scale
				local var2 = ((echoes.twirltimer * 8) << 16) + 0
				A_Custom3DRotate(fl, var1, var2)
				if echoes.abyssactive > 0
					fl.blendmode = AST_SUBTRACT
				else
					fl.frame = $ | FF_ADD
				end
				fl.frame = $ | (fl.fade << 16)
				if echoes.twirltimer <= 12
					fl.fade = $ + 1
				end
				ShieldColorFiller(fl, tracer.player)
			else
				if fl != nil and fl.valid
					P_RemoveMobj(fl)
				end
				table.remove(actor.flames, i)
			end
		end
		if actor.flames2 != nil
			for i, fl in ipairs(actor.flames2) do
				if fl != nil and fl.valid and fl.fade < 10
					local var1 = (0 << 16) + tracer.radius * 3 / tracer.scale
					local var2 = ((echoes.twirltimer * 8) << 16) + 0
					A_Custom3DRotate(fl, var1, var2)
					fl.blendmode = AST_SUBTRACT
					fl.frame = $ | (fl.fade << 16)
					if echoes.twirltimer <= 12
						fl.fade = $ + 1
					end
					ShieldColorFiller(fl, tracer.player)
				else
					if fl != nil and fl.valid
						P_RemoveMobj(fl)
					end
					table.remove(actor.flames2, i)
				end
			end
		end
	else
		OrbitRemover(actor)
	end
end
states[S_ECHOES_FLAMEORBIT] = {
	tics = 1,
	action = A_EchoesFlameOrbit,
	nextstate = S_ECHOES_FLAMEORBIT
}

local function SpinTrigger(player)
	if not SkinCheck(player)
		return
	end
	
	local echoes = player.echoes
	if not echoes.spinpress
		if not FailVibeCheck(player) and not echoes.ropedrop and not
				echoes.shadowleap and not echoes.jumppress and
				echoes.leapcharge <= 0 and echoes.transform < 0 and
				not echoes.voidtumble
			player.pflags = $ & ~PF_SPINNING
			//have void glide orb close much faster
			if echoes.voidorb != nil and echoes.voidorb.valid and not
					echoes.voidlaunch
				echoes.voidorb.fastclose = true
				if echoes.voidcooldown <= 0
					if echoes.voidcooldown == -1
						JumpManager(player.mo, 8 * FRACUNIT)
					end
					echoes.voidcooldown = TICRATE
				end
			//start crouching
			elseif P_GetPlayerControlDirection(player) == 0 and
					FoxOnGround(player.mo)
				echoes.crouch = true
				StateSetter(player.mo, S_ECHOES_CROUCH)
			//trigger phantom twirl and stomp
			elseif echoes.phanttwirl < 2
				player.pflags = $ & ~PF_THOKKED & ~PF_STARTJUMP
				echoes.twirltimer = phant_twirl_maxtime
				//add a buffer for when phantom dashing off the ground
				if FoxOnGround(player.mo)
					echoes.twirltimer = $ + 1
				end
				//and also make the player jump
				if not (player.pflags & PF_JUMPED)
					if FoxOnGround(player.mo)
						P_DoJump(player, false)
					else
						player.pflags = $ | PF_JUMPED | PF_NOJUMPDAMAGE
					end
				end
				if echoes.phanttwirl == 1 and not FoxOnGround(player.mo)
					JumpManager(player.mo, -10 * FRACUNIT)
					S_StartSound(player.mo, sfx_phda3)
				else
					if P_GetPlayerControlDirection(player) != 0
						P_InstaThrust(player.mo, player.drawangle, 
								Skipchoes.skipdive(player))
					end
					local jumpboost = 6
					if echoes.abyssactive <= 0 or IsNerfed()
						jumpboost = $ / 2
					end
					local relative = abs(player.mo.momz) > jumpboost * 
							player.mo.scale / 2 and not isFalling(player.mo)
					JumpManager(player.mo, jumpboost * FRACUNIT, relative)
					if echoes.abyssactive > 0
						S_StartSound(player.mo, sfx_phda2)
					else
						S_StartSound(player.mo, sfx_phda1)
					end
				end
				echoes.phanttwirl = $ + 1
			end
		end
		echoes.spinpress = true
	end
end

//making this a separate function so that it can be used for shadow leap and stomp
local function ExplosionFlames(player, radius)
	if isSuper(player)
		radius = $ * 2
	end
	local landvicts = SearchRadius(player.mo, nil, radius)
	for i, v in ipairs(landvicts) do
		if v != nil and v.valid and v.health and v.state != v.info.deathstate 
				and not isRing(v)
			if v.flags & MF_ENEMY or v.flags & MF_BOSS or
					v.flags & MF_MONITOR
				P_DamageMobj(v, player.mo, player.mo, 1)
			else
				P_KillMobj(v, player.mo, player.mo)
			end
		end
	end
	local ripple = ObjectSpawner(player.mo, MT_ECHOES_EFFECT)
	ripple.state = S_ECHOES_RIPPLE
	ShieldColorFiller(ripple, player)
	ripple.scale = $ * (radius / (125 * FRACUNIT))
	local flames = ObjectSpawner(player.mo, MT_ECHOES_EFFECT)
	flames.tracer = player.mo
	flames.minrad = radius
	flames.stomp = true
	flames.state = S_ECHOES_FLAMEORBIT
	CamShake(player, true)
end

//main thinker for phantom twirl behavior
Echoes.PhantomTwirl = function(player)
	if not SkinCheck(player)
		return
	end
	
	local echoes = player.echoes
	if not (player.cmd.buttons & BT_SPIN) and echoes.spinpress
		echoes.spinpress = false
	end
	if FailVibeCheck(player) or (FoxOnGround(player.mo) and 
			echoes.twirltimer <= phant_twirl_maxtime) or echoes.transform >= 0
			or echoes.voidtumble
		//activate stomp landing
		if echoes.twirltimer > 0 and echoes.phanttwirl == 2 and 
				FoxOnGround(player.mo) and not FailVibeCheck(player) and
				not (player.mo.eflags & MFE_SPRUNG)
			player.mo.momx = $ / 2
			player.mo.momy = $ / 2
			JumpManager(player.mo, 5 * FRACUNIT)
			player.mo.state = S_ECHOES_AIRROLL
			if echoes.abyssactive > 0
				ExplosionFlames(player, 250 * player.mo.scale)
			end
			player.pflags = $ | PF_JUMPED
			S_StartSound(player.mo, sfx_phda4)
		else
			echoes.phanttwirl = 0
		end
		echoes.twirltimer = 0
		StateReset(player.mo, {S_ECHOES_TWIRL})
		return
	end
	
	//handle what happens during twirl states
	local twirlflags = STR_ATTACK | STR_WALL | STR_FLOOR | STR_CEILING | 
			STR_SPIKE
	if echoes.phanttwirl > 1
		twirlflags = $ | STR_SPRING | STR_HEAVY
	end
	if echoes.twirltimer > 0
		StateSetter(player.mo, S_ECHOES_TWIRL, twirlflags)
		if echoes.twirltimer >= phant_twirl_maxtime * 5 / 12
			player.mo.anim_duration = 1
		else
			if leveltime % 2 == 0
				player.mo.anim_duration = 2
			end
		end
		if echoes.abyssactive > 0
			ShieldColorFiller(ParticleSpawner(player.mo), player)
		end
		if echoes.twirlflames == nil or not echoes.twirlflames.valid
			echoes.twirlflames = ObjectSpawner(player.mo, MT_ECHOES_EFFECT)
			echoes.twirlflames.tracer = player.mo
			echoes.twirlflames.state = S_ECHOES_FLAMEORBIT
		end
		if echoes.twirltimer > phant_twirl_maxtime - phant_twirl_maxtime / 3
				and leveltime % 3 == 0
			local ghost = P_SpawnGhostMobj(player.mo)
			if echoes.abyssactive > 0
				ghost.blendmode = AST_MODULATE
			end
		end
		//expand hitbox for super
		if isSuper(player)
			local supervicts = SearchRadius(player.mo, nil, player.mo.radius * 3)
			for i, v in ipairs(supervicts) do
				if v != nil and v.valid and v.health and v.state != 
						v.info.deathstate and not isRing(v)
					if v.flags & MF_ENEMY or v.flags & MF_BOSS or
							v.flags & MF_MONITOR
						P_DamageMobj(v, player.mo, player.mo, 1)
					else
						P_KillMobj(v, player.mo, player.mo)
					end
				end
			end
		end
		if echoes.phanttwirl < 2
			if isFalling(player.mo)
				JumpManager(player.mo, FRACUNIT / 3, true)
			end
			if (echoes.abyssactive <= 0 or IsNerfed()) and leveltime % 2 == 0
				echoes.twirltimer = max($ - 1, 0)
			end
			echoes.twirltimer = max($ - 1, 0)
		else
			//failsafe
			if player.mo.eflags & MFE_SPRUNG or player.mo.eflags & MFE_GOOWATER
					or (player.mo.momz != 0 and not isFalling(player.mo))
				echoes.twirltimer = 0
				return
			end
			if player.mo.momx > 0
				player.mo.momx = max($ - ($ / 10), 0)
			else
				player.mo.momx = min($ - ($ / 10), 0)
			end
			if player.mo.momy > 0
				player.mo.momy = max($ - ($ / 10), 0)
			else
				player.mo.momy = min($ - ($ / 10), 0)
			end
			JumpManager(player.mo, -2 * FRACUNIT, true)
		end
	else
		StateReset(player.mo, {S_ECHOES_TWIRL})
	end
end

//handle what happens when echoes collides against a wall while dashing
Echoes.TwirlBounce = function(mo, thing, line)
	if mo != nil and mo.valid and mo.player != nil and mo.player.valid and
			mo.skin == "echoes" and mo.player.echoes != nil and 
			mo.player.echoes.twirltimer > 0 and mo.player.echoes.phanttwirl < 2
			and line != nil and line.valid
		local echoes = mo.player.echoes
		echoes.twirltimer = 0
		if echoes.abyssactive <= 0
			mo.player.pflags = $ | PF_THOKKED
			echoes.voidtumble = true
			PainSound(mo)
			local BonkSound = {
				sfx_bonk1, sfx_bonk2, sfx_bonk3
			}
			S_StartSound(mo, BonkSound[P_RandomRange(1, 3)])
		else
			JumpManager(mo, 13 * FRACUNIT)
		end
		if isSuper(mo.player)
			echoes.phanttwirl = 0
		end
		local explosion = ObjectSpawner(mo, MT_ECHOES_EFFECT)
		explosion.state = S_CACOFIRE_EXPLODE1
		explosion.colorized = true
		if echoes.abyss_shield != SH_NONE
			ShieldColorFiller(explosion, mo.player)
		else
			explosion.color = ColorOpposite(mo.player.skincolor)
		end
		P_InstaThrust(mo, mo.player.drawangle + ANGLE_180, 12 * mo.scale)
		S_StartSound(mo, sfx_phda6)
		CamShake(mo.player)
	end
end

//handle what happens when echoes hits an enemy while phantom twirling
local function TwirlDamage(target, inflictor)
	if target != nil and target.valid and (target.flags & MF_ENEMY or
			target.flags & MF_MONITOR or target.flags & MF_BOSS) and 
			inflictor != nil and inflictor.valid and inflictor.player != nil 
			and inflictor.player.valid and inflictor.skin == "echoes" and
			inflictor.player.echoes != nil and 
			inflictor.player.echoes.twirltimer > 0 and
			inflictor.player.echoes.phanttwirl < 2
		local echoes = inflictor.player.echoes
		echoes.phanttwirl = 0
		echoes.twirltimer = 0
		if FoxOnGround(inflictor)
			JumpManager(inflictor, 5 * FRACUNIT)
			inflictor.player.pflags = $ | PF_JUMPED
		else
			if inflictor.momz < 0
				inflictor.momz = max($, -10 * inflictor.scale)
			else
				inflictor.momz = min($, 10 * inflictor.scale)
			end
			if not (inflictor.player.pflags & PF_JUMPED)
				inflictor.player.pflags = $ | PF_JUMPED
			end
		end
		local explosion = ObjectSpawner(target, MT_ECHOES_EFFECT)
		explosion.state = S_CACOFIRE_EXPLODE1
		explosion.colorized = true
		if echoes.abyss_shield != SH_NONE
			ShieldColorFiller(explosion, inflictor.player)
		else
			explosion.color = ColorOpposite(inflictor.player.skincolor)
		end
		inflictor.momx = $ / 2
		inflictor.momy = $ / 2
		S_StartSound(inflictor, sfx_phda5)
		CamShake(inflictor.player)
	end
end

//handle what happens during crouch and shadow leap
Echoes.CrouchLeapThinker = function(player)
	if not SkinCheck(player)
		return
	end
	
	local echoes = player.echoes
	if FailVibeCheck(player) or echoes.transform >= 0 or echoes.voidtumble
		player.height = skins[player.mo.skin].height
		echoes.crouch = false
		echoes.shadowleap = false
		echoes.leapcharge = 0
		echoes.chargeheight = 0
		echoes.chargerise = false
		echoes.chargeripple = 0
		echoes.maxleapchrg = false
		StateReset(player.mo, {S_ECHOES_CROUCH, S_PLAY_SPINDASH})
		return
	end
	
	local max_leap_charge = Echoes.max_leap_charge
	if echoes.crouch
		if echoes.spinpress
			player.mo.friction = 32*FRACUNIT/33
			player.pflags = $1 & ~PF_APPLYAUTOBRAKE
			player.pflags = $|PF_STASIS
			if leveltime % 3 == 0 and P_IsObjectOnGround(player.mo) and 
					player.speed > 1*player.mo.scale then 
				P_SpawnSkidDust(player, player.mo.radius, true) 
			end
			
			if echoes.leapcharge <= 0
				StateSetter(player.mo, S_ECHOES_CROUCH)
				player.height = player.spinheight
				if not FoxOnGround(player.mo)
					player.height = skins[player.mo.skin].height
					echoes.crouch = false
				end
			end
		else
			player.height = skins[player.mo.skin].height
			echoes.crouch = false
			echoes.leapcharge = 0
		end
	elseif echoes.leapcharge > 0
		//activate shadow leap once jump is let go
		if not (player.cmd.buttons & BT_JUMP)
			if echoes.leapcharge < 20
				StateSetter(player.mo, S_ECHOES_LEAPCHARGE)
				JumpManager(player.mo, echoes.chargeheight)
				ShieldColorFiller(ParticleSpawner(player.mo), player)
				if FixedHypot(player.mo.momx, player.mo.momy) >= 2*player.mo.scale
					player.powers[pw_nocontrol] = 2 + 32768
					P_Thrust(player.mo, R_PointToAngle2(player.mo.x+player.mo.momx, 
							player.mo.y+player.mo.momy, player.mo.x, player.mo.y), 
							(player.mo.scale*2)/3)
				else
					player.mo.momx = 0 + player.cmomx
					player.mo.momy = 0 + player.cmomy
				end
				echoes.leapcharge = min($ + 1, max_leap_charge)
				if isSuper(player)
					echoes.leapcharge = min($ + 1, max_leap_charge)
				end
				if leveltime % 5 == 0
					CamShake(player)
				end
			else
				local leapboost = echoes.leapcharge * FRACUNIT
				if echoes.leapcharge > 25
					leapboost = $ - echoes.leapcharge * FRACUNIT / 3
				else
					leapboost = $ - echoes.leapcharge * FRACUNIT / 5
				end
				if echoes.abyssactive <= 0 or IsNerfed()
					leapboost = $ - $ / 3
				end
				if P_GetPlayerControlDirection(player) != 0
					P_InstaThrust(player.mo, player.mo.angle, 10 * 
							player.mo.scale)
					if not isSuper(player)
						leapboost = $ - $ / 5
					end
				end
				JumpManager(player.mo, leapboost)
				player.pflags = $ | PF_JUMPED | PF_NOJUMPDAMAGE
				echoes.shadowleap = true
				local radius = 250 * player.mo.scale
				if echoes.leapcharge > 20 and (echoes.abyssactive > 0 and
						not IsNerfed())
					radius = $ * (echoes.leapcharge / 20)
				end
				ExplosionFlames(player, radius)
				if echoes.leapcharge >= max_leap_charge
					S_StartSound(player.mo, sfx_shlp4)
				else
					S_StartSound(player.mo, sfx_shlp3)
				end
				echoes.leapcharge = 0
				echoes.chargeheight = 0
				echoes.chargerise = false
				echoes.chargeripple = 0
			end
		end
	elseif echoes.shadowleap
		local powflags
		if echoes.abyssactive > 0
			local particle = ParticleSpawner(player.mo)
			if echoes.maxleapchrg
				particle.sprite = SPR_PTSP
			end
			ShieldColorFiller(particle, player)
			powflags = STR_ATTACK
		end
		StateSetter(player.mo, S_PLAY_SPINDASH, powflags)
		if abs(player.mo.momz) > 15 * player.mo.scale
			player.mo.tics = 1
		else
			if leveltime % 3 == 0
				player.mo.tics = 3
			end
		end
		if isFalling(player.mo) or FoxOnGround(player.mo)
			echoes.shadowleap = false
			echoes.maxleapchrg = false
		elseif abs(player.mo.momz) > 10 * player.mo.scale
			local ghost = P_SpawnGhostMobj(player.mo)
			if echoes.abyssactive > 0
				ghost.blendmode = AST_MODULATE
			end
		end
	else
		StateReset(player.mo, {S_ECHOES_CROUCH, S_PLAY_SPINDASH})
	end
end

//hooks
addHook("SpinSpecial", SpinTrigger)
addHook("MobjDamage", TwirlDamage)
addHook("MobjDeath", TwirlDamage)
//quick hook for taking ziplines into account
addHook("PreThinkFrame", function()
	for player in players.iterate do
		if not SkinCheck(player)
			continue
		end
		
		local echoes = player.echoes
		if player.cmd.buttons & BT_SPIN and player.powers[pw_carry] ==
				CR_ROPEHANG
			echoes.ropedrop = true
		elseif (FailVibeCheck(player) or FoxOnGround(player.mo)
				or not (player.cmd.buttons & BT_SPIN)) and echoes.ropedrop
			echoes.ropedrop = false
		end
	end
end)