//Echoes' animation stuff, coded by Emmette Clyde

local SkinCheck = Echoes.SkinCheck
local FailVibeCheck = Echoes.FailVibeCheck
local JumpManager = Echoes.JumpManager
local ParticleSpawner = Echoes.ParticleSpawner
local ShieldColorFiller = Echoes.ShieldColorFiller
local ObjectSpawner = Echoes.ObjectSpawner
local isSuper = Echoes.isSuper
local StateSetter = Echoes.StateSetter
local FoxOnGround = Echoes.FoxOnGround
local CamShake = Echoes.CamShake
local SpeedCalc = Echoes.SpeedCalc
local StateReset = Echoes.StateReset
local ObjectZCoord = Echoes.ObjectZCoord
local isFalling = Echoes.isFalling
local GetFloorZ = Echoes.GetFloorZ
local isRing = Echoes.isRing
local SearchRadius = Echoes.SearchRadius
local PlayerCam = Echoes.PlayerCam

//custom action for spawning emeralds in and having them orbit during transformation
local function EmeraldRemover(mo)
	if mo != nil and mo.valid
		mo.removing = true
		if mo.echemerls != nil
			for i, em in ipairs(mo.echemerls) do
				if em != nil and em.valid
					P_RemoveMobj(em)
				end
				table.remove(mo.echemerls, i)
			end
		end
		if mo.echemerls == nil or unpack(mo.echemerls) == ""
			P_RemoveMobj(mo)
		end
	end
end

function A_EchoesEmeralds(actor, var1, var2)
	if actor == nil or not actor.valid
		return
	end
	if actor.removing
		if actor.echemerls != nil
			for i, em in ipairs(actor.echemerls) do
				if em != nil and em.valid
					em.state = S_INVISIBLE
				end
			end
		end
		EmeraldRemover(actor)
		return
	end
	local tracer = actor.tracer
	if tracer == nil or not tracer.valid or tracer.player == nil or not
			tracer.player.valid or tracer.skin != "echoes" or 
			tracer.player.echoes == nil
		EmeraldRemover(actor)
		return
	end
	local echoes = tracer.player.echoes
	if echoes.transform > 0
		P_MoveOrigin(actor, tracer.x, tracer.y, ObjectZCoord(tracer, actor, 
				tracer.height / 2))
		if actor.echemerls == nil
			actor.echemerls = {}
		end
		local emcount = 0
		for i, em in ipairs(actor.echemerls) do
			if em != nil and em.valid
				emcount = $ + 1
			end
		end
		if echoes.transform % (TICRATE / 3) == 0 and emcount < 7
			local EchoEmerls = {
				S_CEMG1, S_CEMG2, S_CEMG3, S_CEMG4, S_CEMG5, S_CEMG6, S_CEMG7
			}
			local emerl = ObjectSpawner(actor, MT_ECHOES_EFFECT)
			emerl.state = EchoEmerls[emcount + 1]
			emerl.angle = actor.angle
			emerl.orbrad = actor.radius
			emerl.target = actor
			table.insert(actor.echemerls, emerl)
		end
		for i, em in ipairs(actor.echemerls) do
			if em != nil and em.valid
				em.orbrad = min($ + actor.scale * 7, 75 * actor.scale)
				local var1 = (0 << 16) + em.orbrad / actor.scale
				local var2
				if emcount >= 7
					var2 = ((2000 / echoes.transform) << 16)
				else
					var2 = (45 << 16)
				end
				A_Custom3DRotate(em, var1, var2 + 0)
			else
				table.remove(actor.echemerls, i)
			end
		end
	else
		EmeraldRemover(actor)
	end
end
states[S_ECHOES_ORBEMERLS] = {
	tics = 1,
	action = A_EchoesEmeralds,
	nextstate = S_ECHOES_ORBEMERLS
}

//handle the transformation sequence
Echoes.SoulFusionTransform = function(player)
	if not SkinCheck(player)
		return
	end
	
	local echoes = player.echoes
	//failsafe for if echoes is no longer super
	if (echoes.abyssactive > 2 or player.charflags & SF_SUPER) and not 
			isSuper(player) and echoes.transform < 0
		player.charflags = $ & ~SF_SUPER
		echoes.abyssactive = 1
	end
	//prevent transforming if we can't
	if FailVibeCheck(player) or echoes.voidtumble
		echoes.transform = -1
		return
	end
	
	local cam = PlayerCam(player)
	//trigger the start of the transformation
	if player.cmd.buttons & BT_CUSTOM2 and All7Emeralds(emeralds) and 
			player.rings >= 50 and player.pflags & PF_JUMPED and not 
			isSuper(player) and echoes.transform < 0
		player.charflags = $ | SF_SUPER
		echoes.transform = 3 * TICRATE
		echoes.abyssactive = 3
		player.powers[pw_shield] = SH_NONE
		echoes.abyss_shield = SH_NONE
		//spawn in objects that will become emeralds
		local emorbiter = ObjectSpawner(player.mo, MT_ECHOES_EFFECT)
		emorbiter.tracer = player.mo
		emorbiter.state = S_ECHOES_ORBEMERLS
		emorbiter.angle = cam.angle + ANGLE_180
		S_FadeOutStopMusic(3 * MUSICRATE, player)
		S_StartSound(player.mo, sfx_prme1)
	end
	//handle the actual sequence
	if echoes.transform > 0
		StateSetter(player.mo, S_ECHOES_LEAPCHARGE, STR_GUARD)
		player.drawangle = cam.angle + ANGLE_180
		A_ForceStop(player.mo)
		player.pflags = $ | PF_FULLSTASIS
		JumpManager(player.mo, FRACUNIT / 3)
		player.charflags = $ | SF_SUPER
		if echoes.abyss != nil and echoes.abyss.valid
			local abyss = echoes.abyss
			abyss.target = player.mo
			StateSetter(abyss, S_PLAY_NIGHTS_TRANS1)
			abyss.frame = 0 | ($ & ~FF_FRAMEMASK)
			local var2 = ((5500 / echoes.transform) << 16) + 0
			A_Custom3DRotate(abyss, (0 << 16) + echoes.transform, var2)
		end
		if echoes.transform % 20 == 0
			local ripple = ObjectSpawner(player.mo, MT_ECHOES_EFFECT, 0, 0, 
					player.mo.height / 2)
			ripple.state = S_ECHOES_RIPPLE
			ripple.sprite = SPR_ECCR
			ripple.dispoffset = -1
			ShieldColorFiller(ripple, player)
			ripple.scale = $ * 3
		end
		echoes.transform = $ - 1
	//handle what happens at the end of the sequence
	elseif echoes.transform == 0
		local transvicts = SearchRadius(player.mo, nil, 1000 * player.mo.scale)
		for i, v in ipairs(transvicts) do
			if v != nil and v.valid and v.health and v.state != 
					v.info.deathstate and not isRing(v)
				if v.flags & MF_ENEMY or v.flags & MF_BOSS or v.flags & 
						MF_MONITOR
					P_DamageMobj(v, player.mo, player.mo, 1)
				else
					P_KillMobj(v, player.mo, player.mo)
				end
			end
		end
		P_FlashPal(player, PAL_WHITE, 5)
		player.charflags = $ | SF_SUPER
		P_DoSuperTransformation(player, false)
		CamShake(player, true)
		S_StartSound(player.mo, sfx_prme2)
		echoes.transform = $ - 1
	end
end

//some misc animations that don't impact gameplay
Echoes.MiscAnimations = function(player)
	if not SkinCheck(player)
		return
	end
	
	//clear skip's stuff from echoes when she fails the vibe check
	if FailVibeCheck(player)
		if player.skipsmug
			player.skipsmug = false
			if player.playerstate != PST_DEAD
				P_RestoreMusic(player)
			end
		end
		player.mo.skipsmugtimer = 0
	end
	
	/*failsafes because of that one vanilla bug where jumping in a spin gap 
	bugs the jump & spin states*/
	if player.pflags & PF_SPINNING and player.mo.state != S_PLAY_ROLL
		player.mo.state = S_PLAY_ROLL
	end
	if P_IsObjectOnGround(player.mo) and (player.pflags & PF_JUMPED or 
			player.pflags & PF_THOKKED)
		player.pflags = $ & ~PF_JUMPED & ~PF_THOKKED
	end
	
	local echoes = player.echoes
	//spawn flame trail for echoes while she's running
	if player.mo.state == S_PLAY_RUN
		local runfire = ParticleSpawner(player.mo)
		runfire.sprite = SPR_PFLM
		P_SetOrigin(runfire, runfire.x, runfire.y, player.mo.z)
		ShieldColorFiller(runfire, player)
	end
	//have particles spawn around echoes while super
	if isSuper(player)
		local particle = ParticleSpawner(player.mo)
		particle.sprite = SPR_PFLM
		ShieldColorFiller(particle, player)
	end
	//change the super theme to the prime theme
	if player.powers[pw_super]
		if not echoes.primusic
			P_PlayJingleMusic(player, "PRIME", 0, true, JT_SUPER)
			echoes.primusic = true
		end
		//and also give her a "floating" effect while on the ground
		if player.mo.state != S_PLAY_EDGE and player.mo.state != S_ECHOES_CROUCH
				and FoxOnGround(player.mo) and not player.skipsmug and
				player.mo.state != S_PLAY_FIRE and player.mo.state != 
				S_PLAY_FIRE_FINISH and not echoes.outoftouch
			if echoes.orioffset == nil
				echoes.orioffset = player.mo.spriteyoffset
			end
			player.mo.spriteyoffset = $ + echoes.primflot
			if leveltime % 2 == 0
				if echoes.primascend
					echoes.primflot = $ + FRACUNIT / 5
					if echoes.primflot > FRACUNIT / 2
						echoes.primascend = false
					end
				else
					echoes.primflot = $ - FRACUNIT / 5
					if echoes.primflot < FRACUNIT / 2 * -1
						echoes.primascend = true
					end
				end
			end
		elseif player.mo.spriteyoffset != 0
			player.mo.spriteyoffset = 0
		end
	else
		if echoes.primusic
			echoes.primusic = false
		end
		if player.mo.spriteyoffset != 0
			player.mo.spriteyoffset = 0
		end
	end
	//change her standing and waiting states while vulnerable
	if echoes.abyssactive <= 0 and (player.mo.state == S_PLAY_STND or
			player.mo.state == S_PLAY_WAIT) and not FailVibeCheck(player)
		player.mo.state = S_PLAY_CLING
		player.mo.anim_duration = 15
	elseif (echoes.abyssactive > 0 or SpeedCalc(player) > 0 or not 
			FoxOnGround(player.mo) or FailVibeCheck(player)) and 
			player.mo.state == S_PLAY_CLING
		StateReset(player.mo, {S_PLAY_CLING})
	elseif echoes.abyssactive <= 0 and player.mo.state == S_PLAY_CLING
		if leveltime % 15 == 0
			player.mo.anim_duration = 15
		end
	end
	//add an indicator for the stun cooldown
	if echoes.stuncooldown > 0
		local z = ObjectZCoord(player.mo, echoes.stunindi, player.mo.height +
				player.mo.height * 3 / 4)
		if echoes.stunindi == nil or not echoes.stunindi.valid
			echoes.stunindi = ObjectSpawner(player.mo, MT_ECHOES_EFFECT)
			echoes.stunindi.state = S_ECHOES_WARNING
			echoes.stunindi.color = player.skincolor
			P_SetOrigin(echoes.stunindi, player.mo.x, player.mo.y, z)
		else
			P_MoveOrigin(echoes.stunindi, player.mo.x, player.mo.y, z)
		end
	elseif echoes.stunindi != nil and echoes.stunindi.valid
		P_RemoveMobj(echoes.stunindi)
		echoes.stunindi = nil
	end
	//toggle having abyss visible or not
	if player.cmd.buttons & BT_CUSTOM1 and not echoes.cust1press
		echoes.hideabyss = not $
		echoes.cust1press = true
	elseif not (player.cmd.buttons & BT_CUSTOM1) and echoes.cust1press
		echoes.cust1press = false
	end
	if echoes.hideabyss and echoes.abyssactive > 0 and not isSuper(player) and
			(echoes.voidorb == nil or not echoes.voidorb.valid) and 
			echoes.twirltimer <= 0 and echoes.leapcharge <= 0 and not 
			echoes.shadowleap and (player.powers[pw_carry] != 19 or 
			not player.mo.boat)
		ShieldColorFiller(ParticleSpawner(player.mo), player)
	end
	//remove the highres flag from echoes while she's hugging skip
	if player.mo.state == S_PLAY_LOVERS_HUG
		//gotta use a player var to prevent weird imcompatibilities with anims
		if not echoes.skiphug
			player.mo.spritexscale = FRACUNIT + (FRACUNIT * 5 / 100)
			player.mo.spriteyscale = FRACUNIT + (FRACUNIT * 5 / 100)
			echoes.skiphug = true
		end
	elseif echoes.skiphug
		player.mo.spritexscale = FRACUNIT
		player.mo.spriteyscale = FRACUNIT
		echoes.skiphug = false
	end
	//handle the air animations
	if player.pflags & PF_JUMPED and (echoes.voidorb == nil or not
			echoes.voidorb.valid) and echoes.twirltimer <= 0
			and echoes.leapcharge <= 0 and not echoes.shadowleap
			and (isFalling(player.mo) or abs(player.mo.momz) <= FRACUNIT)
			and (player.mo.state == S_PLAY_JUMP or 
			player.mo.state == S_PLAY_SPRING)
		player.mo.state = S_ECHOES_AIRROLL
		player.pflags = $ | PF_JUMPED
	end
end

//let's trigger the death animation, yay
local function DeathTrigger(target, inflictor, source, damagetype)
	if target != nil and target.valid and target.player != nil and 
			target.player.valid and target.skin == "echoes" and
			target.player.echoes != nil
		local echoes = target.player.echoes
		if damagetype != nil
			if damagetype == DMG_DEATHPIT
				echoes.deathtype = 2
			else
				echoes.deathtype = 1
			end
		else
			echoes.deathtype = 1
		end
		if echoes.deathtype == 2
			S_StartSound(target, sfx_eded3)
		else
			S_StartSound(target, sfx_eded1)
		end
	end
end

Echoes.DeathThinker = function(player)
	if not SkinCheck(player)
		return
	end
	
	local echoes = player.echoes
	if player.playerstate != PST_DEAD
		echoes.deathtype = -1
		echoes.dedstart = false
		return
	elseif echoes.deathtype >= 0
		if echoes.deathtype == 2
			StateSetter(player.mo, S_PLAY_STUN)
			JumpManager(player.mo, -30 * FRACUNIT)
		else
			player.mo.flags = $ & ~MF_NOCLIP & ~MF_NOCLIPHEIGHT
			player.mo.fuse = 5
			if not echoes.dedstart
				player.mo.state = S_ECHOES_DEATH1
				S_StartSound(player.mo, sfx_eded2)
				echoes.dedstart = true
			end
			player.mo.momz = 0
		end
		if player.deadtimer == 70 and player.lives <= 0
			P_PlayJingleMusic(player, "EDED4", 0, false, JT_GOVER)
		end
	end
end

//hooks
addHook("MobjDeath", DeathTrigger, MT_PLAYER)
//trigger funny "fun" voiceline when echoes says fun in chat
addHook("PlayerMsg", function(source, msgtype, target, msg)
	if SkinCheck(source) and msg != nil and msg != "" and msgtype == 0
		msg = string.upper($)
		msg = string.gsub($, "[%p%c%s]", "")
		if string.find(msg, "FUN")
			S_StartSound(source.mo, sfx_efun)
		end
	end
end)