-- EveryoneHugs detection.
-- Just do your think normally if it's not added, but let it take control once it is.
local everyonehugs = false

addHook("ThinkFrame", do
	if AddHugger and not everyonehugs
		AddHugger("skip")
		everyonehugs = true
	end
end)

addHook("NetVars", function(net)
	everyonehugs = net($)
end)

-- Heart object definition
freeslot("S_PLAY_SKIPHUG", "S_COLORHEART", "SPR2_SHUG", "SPR_RHRT")

states[S_PLAY_SKIPHUG] = {SPR_PLAY, SPR2_SHUG|FF_ANIMATE, 12, nil, 3, 4, S_PLAY_SKIPHUG}

states[S_COLORHEART] = {
	sprite = SPR_RHRT,
	tics = -1
}

-- Love is in the air!
local function SpawnHearts(player)
	local heart = P_SpawnMobjFromMobj(player.mo, 0, 0, player.mo.height, MT_CDLHRT)
	heart.destscale = (5*player.mo.scale)>>4
	heart.fuse = (5*TICRATE)>>1
	heart.momz = player.mo.scale
	heart.target = player.mo
	heart.extravalue1 = player.mo.x
	heart.extravalue2 = player.mo.y
end

-- Let's find the closest person to hug!
local function FindHugCandidates(rosy)
	local huggable
	
	for p in players.iterate
		-- Don't hug yourself
		if p == rosy continue end
		
		-- Hug only living players
		if not p.mo continue end
		if p.playerstate != PST_LIVE continue end
		
		-- No hugging in NiGHTS
		if p.powers[pw_carry] == CR_NIGHTSMODE continue end
		
		-- It's another Rosy and she's not hugging you
		if p.StephChars_hugging and rosy.StephChars_hugged != p continue end
		
		-- No hugging the enemy
		if p.ctfteam != rosy.ctfteam continue end
		
		-- Can't hug if not on the floor or doing anything
		if not(P_IsObjectOnGround(p.mo) and (p.panim == PA_IDLE or p.panim == PA_WALK or p.panim == PA_RUN))
			continue
		end
		
		-- Too high or too low to hug
		if abs(rosy.mo.z - p.mo.z) > 4*FRACUNIT continue end
		
		-- Too far away
		local dist = R_PointToDist2(rosy.mo.x, rosy.mo.y, p.mo.x, p.mo.y)
		if dist > rosy.mo.radius + p.mo.radius continue end
		
		-- Update the most likely candidate
		if not huggable or (huggable and dist < huggable.dist)
			huggable = p
			huggable.dist = dist
		end
	end
	
	return huggable
end

Skipchoes.ReleaseHug = function(p)
	if p.StephChars_hugging -- We're hugging
		-- Do this to me
		p.pflags = $ & ~PF_STASIS
		if P_IsObjectOnGround(p.mo) then p.mo.state = S_PLAY_STND end
		p.FAV_Hugbacks = nil
		p.StephChars_hugdelay = TICRATE
		
		-- Do this to them
		if p.StephChars_hugging.valid
			p.StephChars_hugging.pflags = $ & ~PF_STASIS
			p.StephChars_hugging.StephChars_hugged = nil
			p.StephChars_hugging.hugged = nil
			if p.StephChars_hugging.mo.state == S_PLAY_LOVERS_HUG
				if P_IsObjectOnGround(p.StephChars_hugging.mo) then p.StephChars_hugging.mo.state = S_PLAY_STND end
			end
		end
		
		p.StephChars_hugging = nil
		p.hugging = nil
	elseif p.StephChars_hugged -- We're being hugged
		-- Do this for me
		p.pflags = $ & ~PF_STASIS
		if p.mo.state == S_PLAY_LOVERS_HUG
			if P_IsObjectOnGround(p.mo) then p.mo.state = S_PLAY_STND end
		end
		p.FAV_Hugbacks = nil
		
		-- Do this to them
		if p.StephChars_hugged.valid
			p.StephChars_hugged.pflags = $ & ~PF_STASIS
			--p.StephChars_hugged.mo.angle = R_PointToAngle2(p.mo.x, p.mo.y, p.StephChars_hugged.mo.x, p.StephChars_hugged.mo.y)
			p.StephChars_hugged.drawangle = p.StephChars_hugged.mo.angle
			p.StephChars_hugged.mo.state = S_PLAY_PAIN
			
			if p.StephChars_hugged.StephChars_hugtime > 1*TICRATE
				-- That's enough hugging
				p.StephChars_hugged.mo.sprite2 = SPR2_JUMP
				p.StephChars_hugged.mo.frame = A
				p.StephChars_hugged.thrown = true
			else
				-- Ouch!
				p.StephChars_hugged.mo.sprite2 = SPR2_PAIN
				S_StartSound(p.StephChars_hugged.mo, sfx_cdpcm7)
			end
			
			-- Fling Amy
			A_DoNPCPain(p.StephChars_hugged.mo)
			
			p.StephChars_hugged.StephChars_hugging = nil
			p.StephChars_hugged.hugging = nil
			p.StephChars_hugged.FAV_Hugbacks = nil
		end
		
		p.StephChars_hugged = nil
		p.hugged = nil
		p.StephChars_hugdelay = TICRATE
	end
end
local ReleaseHug = Skipchoes.ReleaseHug

-- Release!
local function UnhugEveryone()
	for p in players.iterate
		if p.StephChars_hug
			ReleaseHug(p)
		end
	end
end

freeslot("S_PLAY_LOVERS_HUG")
states[S_PLAY_LOVERS_HUG] = {
	sprite = SPR_PLAY,
	frame = SPR2_TWIN | FF_ANIMATE | A,
	tics = 24,
	nextstate = S_PLAY_LOVERS_HUG,
	var1 = 4,
	var2 = 6
}

addHook("PlayerThink", function(p)
	if not p.mo return end
	
	if p.mo.skin != "skip" and p.mo.skin != "echoes"
		if p.StephChars_hug or p.StephChars_hugging ReleaseHug(p) end
		
		return
	end
	
	//So other functions that rely on these variables can work!
	p.hugged = p.StephChars_hugged
	p.hugging = p.StephChars_hugging
	if not P_IsObjectOnGround(p.mo) then ReleaseHug(p) end
	
	local rosy = p
	
	-- Hug delay ticker
	rosy.StephChars_hugdelay = $ and $ - 1 or 0
	
	-- Curl when you're falling from being thrown
	if rosy.thrown and rosy.mo.momz < 0
		rosy.mo.sprite2 = SPR2_ROLL
		rosy.mo.frame = D
		rosy.thrown = false
	end
	
	-- Hug thinker ~
	if rosy.StephChars_hugging
		if not rosy.StephChars_hugging.valid
			ReleaseHug(rosy)
			return
		end
		
		// hey, you should be reacting to this too. bitch
		rosy.StephChars_hugging.hugged = rosy.StephChars_hugging.StephChars_hugged
		rosy.StephChars_hugging.hugging = rosy.StephChars_hugging.StephChars_hugging
		
		rosy.StephChars_hugging.pflags = $ | PF_STASIS	
		rosy.pflags = $ | PF_STASIS
		rosy.target = rosy.StephChars_hugging
		rosy.drawangle = rosy.StephChars_hugangle
		if rosy.mo.state ~= S_PLAY_SKIPHUG and rosy.StephChars_hugging.mo.skin != "echoes"
				and rosy.mo.skin != "echoes"
			rosy.mo.state = S_PLAY_SKIPHUG
		end
		rosy.StephChars_hugtime = $ + 1
		
		if  rosy.StephChars_hugging.mo.skin == "skip" --Hug back! This is a war of affection!
		and (rosy.StephChars_hugging.cmd.buttons & BT_CUSTOM2) and rosy.mo.skin != "echoes"
			if not (rosy.StephChars_hugging.mo.state == S_PLAY_SKIPHUG)	
					rosy.StephChars_hugged = rosy.StephChars_hugging
					rosy.StephChars_hugging.StephChars_hugging = rosy
					rosy.StephChars_hugging.StephChars_hugtime = 0
					rosy.StephChars_hugging.StephChars_hugangle = R_PointToAngle2(rosy.StephChars_hugging.mo.x, rosy.StephChars_hugging.mo.y, rosy.mo.x, rosy.mo.y)
					rosy.StephChars_hugging.StephChars_hugdelay = TICRATE
					S_StartSound(rosy.StephChars_hugging.mo, sfx_cdpcm6)
			end
		end	
		
		//have skip and echoes display unique hug sprites
		local skip, echoes
		if rosy.mo.skin == "skip"
			skip = rosy
		elseif rosy.StephChars_hugging.mo.skin == "skip"
			skip = rosy.StephChars_hugging
		end
		if rosy.mo.skin == "echoes"
			echoes = rosy
		elseif rosy.StephChars_hugging.mo.skin == "echoes"
			echoes = rosy.StephChars_hugging
		end
		if skip != nil and echoes != nil
			if skip.mo.state != S_PLAY_LOVERS_HUG
				skip.mo.state = S_PLAY_LOVERS_HUG
			end
			if echoes.mo.state != S_PLAY_LOVERS_HUG
				echoes.mo.state = S_PLAY_LOVERS_HUG
				P_MoveOrigin(echoes.mo, skip.mo.x, skip.mo.y, skip.mo.z)
			end
		end
		
		if rosy.cmd.buttons & BT_CUSTOM2 and not rosy.StephChars_hugdelay
			ReleaseHug(rosy)
		elseif rosy.cmd.buttons & BT_CUSTOM3
			if not (leveltime & 7)
				SpawnHearts(rosy)
			end
		end
		
		//failsafe for if the player echoes is hugging is no longer skip
		if rosy.mo.skin == "echoes" and rosy.StephChars_hugging and rosy.StephChars_hugging.mo.skin != "skip"
			ReleaseHug(rosy)
		end
		
		return
	end
	
	local cand = FindHugCandidates(rosy)
	
	//if we're playing as echoes, remove candidate if they aren't skip
	if p.mo.skin == "echoes" and cand != nil and cand.valid and 
			cand.mo.skin != "skip"
		cand = nil
	end
	
	-- We don't have candidates? Get rid of the heart
	if not cand
		if rosy.heart and rosy.heart.valid
			P_RemoveMobj(rosy.heart)
		end
		
		rosy.heart = nil
		return
	end
	
	-- Found someone? Place the heart indicator
	if not rosy.heart
		rosy.heart = P_SpawnMobjFromMobj(cand.mo, 0, 0, cand.mo.height + 24*cand.mo.scale, MT_LOCKON)
		rosy.heart.state = S_COLORHEART
		rosy.heart.color = rosy.mo.color
		rosy.heart.scale = 2*cand.mo.scale/3
		rosy.heart.fuse = -1
		rosy.heart.tracer = rosy.mo
		rosy.heart.target = cand.mo
	end
	
	-- Hug!
	if rosy.cmd.buttons & BT_CUSTOM2 and not rosy.StephChars_hugdelay
	and not (rosy.pflags & PF_STASIS)
	and not (rosy.pflags & PF_FULLSTASIS)
	or rosy.FAV_Hugbacks -- hugbacks!!
		rosy.mo.momx = 0
		rosy.mo.momy = 0
		
		cand.StephChars_hugged = rosy
		if not (rosy.mo.skin == "echoes" and cand.mo.skin == "skip" or rosy.mo.skin == "skip" and cand.mo.skin == "echoes")
			cand.FAV_Hugbacks = rosy
		end
		
		rosy.StephChars_hugging = cand
		
		rosy.StephChars_hugtime = 0
		rosy.StephChars_hugangle = R_PointToAngle2(rosy.mo.x, rosy.mo.y, cand.mo.x, cand.mo.y)
		rosy.StephChars_hugdelay = TICRATE
		
		-- Ufufu~
		S_StartSound(rosy.mo, sfx_cdpcm6)
	end
end)

addHook("MobjThinker", function(mo)
	if not (mo and mo.valid) return end
	if mo and mo.tracer and mo.valid and mo.tracer.valid
		if mo.tracer.skin != "skip" or mo.tracer.player.StephChars_hugging
			mo.tracer.heart = nil
			P_RemoveMobj(mo)
			return
		else
			mo.color = mo.tracer.color
			mo.scale = 2*mo.target.scale/3
			
			local bob = 5 * sin(FixedAngle(leveltime*FRACUNIT*3))
			
			P_MoveOrigin(mo, mo.target.x, mo.target.y, mo.target.z + mo.target.height + 15 * mo.scale + FixedMul(bob, mo.scale))
			return true
		end
	elseif mo.state == S_COLORHEART
		P_RemoveMobj(mo)
	end
end, MT_LOCKON)

addHook("JumpSpecial", ReleaseHug)